# %%
import struct
from enum import Enum
from PIL import Image
import lzma
import os
import numpy as np
from glob import glob
from multiprocessing import Pool
import tqdm
import sys


# %%
DTX_COMMANDSTRING_LEN = 128

# 当前允许的最大纹理大小。
MAX_DTX_SIZE = 1024

# 纹理中最大允许的多级渐远图数量。
MAX_DTX_MIPMAPS = 8

DTX_FULLBRITE = 1 << 0  # 此DTX具有全亮度颜色。
DTX_PREFER16BIT = 1 << 1  # 即使在32位模式下，也使用16位。
DTX_MIPSALLOCED = 1 << 2  # 用于简化一些工具操作。
# 这意味着每个TextureMipData都有其纹理数据分配。
DTX_SECTIONSFIXED = 1 << 3  # 原本的部分计数是错误的。当计数固定时，此标志将设置在所有纹理上。
DTX_NOSYSCACHE = 1 << 6  # 告诉它不要将纹理放入纹理缓存列表中。
DTX_PREFER4444 = 1 << 7  # 如果在16位模式下，使用4444纹理。

DTX_PREFER5551 = 1 << 8  # 如果是16位，使用5551。

DTX_32BITSYSCOPY = 1 << 9  # 如果有系统复制，不要将其转换为设备特定格式（保持32位）。

DTX_CUBEMAP = 1 << 10  # 立方环境贴图。+x存储在正常数据区域中，
# -x,+y,-y,+z,-z存储在其自己的部分中

DTX_BUMPMAP = 1 << 11  # 凹凸贴图纹理，此纹理具有8位U和V分量用于凹凸法线

DTX_LUMBUMPMAP = 1 << 12  # 具有亮度的凹凸贴图纹理，此纹理具有8位亮度，U和V

# 实际保存哪些标志。
DTX_FLAGSAVEMASK = DTX_FULLBRITE | DTX_32BITSYSCOPY | DTX_PREFER16BIT | \
                    DTX_SECTIONSFIXED | DTX_PREFER4444 | DTX_PREFER5551 | \
                    DTX_CUBEMAP | DTX_BUMPMAP | DTX_LUMBUMPMAP | DTX_NOSYSCACHE

CURRENT_DTX_VERSION = -5  # DTX头中的m_Version。

class BPPIdent(Enum):
    BPP_8P = 0  # 8 bit palettized
    BPP_8 = 1  # 8 bit RGB
    BPP_16 = 2
    BPP_32 = 3
    BPP_S3TC_DXT1 = 4
    BPP_S3TC_DXT3 = 5
    BPP_S3TC_DXT5 = 6
    BPP_32P = 7  # this was added for true color palette support
    BPP_24 = 8

# %%
class DtxHeader:
    # Accessors.
    def __init__(self, hdr: bytes):
        # Define the structure format
        # Note: 'I' for uint32, 'i' for int32, 'H' for uint16, 's' for string
        # The union is not directly represented in struct format, so we handle it separately
        struct_format = 'IiHHHHii12s128s'

        # Unpack the data
        unpacked_data = struct.unpack(struct_format, hdr)

        # Extract the values
        self.m_ResType, \
        self.m_Version, \
        self.m_BaseWidth, \
        self.m_BaseHeight, \
        self.m_nMipmaps, \
        self.m_nSections, \
        self.m_IFlags, \
        self.m_UserFlags, \
        self.m_Extra, \
        self.m_CommandString = unpacked_data

        # Convert the extra data to a list of uint8
        self.m_Extra = list(self.m_Extra)
        # Convert the command string to a string
        # self.m_CommandString = self.m_CommandString.decode('utf-8')

    def get_texture_group(self):
        return self.m_Extra[0]

    def get_num_mipmaps(self):
        return self.m_Extra[1]

    def get_bpp_ident(self):
        return self.m_Extra[2] if self.m_Extra[2] != 0 else BPPIdent.BPP_32

    def set_bpp_ident(self, id):
        self.m_Extra[2] = id

    def get_non_s3tc_mipmap_offset(self):
        return self.m_Extra[3]

    def set_non_s3tc_mipmap_offset(self, offset):
        self.m_Extra[3] = offset

    def get_ui_mipmap_offset(self):
        return self.m_Extra[4]

    def get_ui_mipmap_scale(self):
        return 1 << self.get_ui_mipmap_offset()

    def set_ui_mipmap_offset(self, val):
        self.m_Extra[4] = val

    def get_texture_priority(self):
        return self.m_Extra[5]

    def set_texture_priority(self, val):
        self.m_Extra[5] = val
 
    def get_detail_texture_scale(self):
        return struct.unpack('f', struct.pack('I', self.m_Extra[6] | (self.m_Extra[7] << 8) | (self.m_Extra[8] << 16) | (self.m_Extra[9] << 24)))[0] + 1.0

    def set_detail_texture_scale(self, val):
        packed = struct.pack('f', val - 1.0)
        self.m_Extra[6] = packed[0]
        self.m_Extra[7] = packed[1]
        self.m_Extra[8] = packed[2]
        self.m_Extra[9] = packed[3]

    def get_detail_texture_angle(self):
        return self.m_Extra[10] + (self.m_Extra[11] << 8)

    def set_detail_texture_angle(self, angle):
        self.m_Extra[10] = angle & 0xFF
        self.m_Extra[11] = angle >> 8

    def __str__(self) -> str:
        return "\n".join([
               f"ResType: {self.m_ResType}",
               f"Version: {self.m_Version}",
               f"BaseWidth: {self.m_BaseWidth}",
               f"BaseHeight: {self.m_BaseHeight}",
               f"nMipmaps: {self.m_nMipmaps}",
               f"nSections: {self.m_nSections}",
               f"IFlags: {self.m_IFlags:032b}",
               f"UserFlags: {self.m_UserFlags:032b}",
               f"CommandString: \"{self.m_CommandString}\"",
               # Extra
               f"Texture Group: {self.get_texture_group()}",
               f"Num Mipmaps: {self.get_num_mipmaps()}",
               f"BPP Ident: {BPPIdent(self.get_bpp_ident())}",
               f"Non S3TC Mipmap Offset: {self.get_non_s3tc_mipmap_offset()}",
               f"UI Mipmap Offset: {self.get_ui_mipmap_offset()}",
               f"Texture Priority: {self.get_texture_priority()}",
               f"Detail Texture Scale: {self.get_detail_texture_scale()}",
               f"Detail Texture Angle: {self.get_detail_texture_angle()}",
        ])

# %%
def rgb565_to_rgba(rgb565):
    r = ((rgb565 >> 11) & 0x1F) << 3  # Extract and scale red component (5 bits to 8)
    g = ((rgb565 >> 5) & 0x3F) << 2   # Extract and scale green component (6 bits to 8)
    b = (rgb565 & 0x1F) << 3          # Extract and scale blue component (5 bits to 8)
    return np.array((r, g, b, 255))  # Alpha is always 255 for DXT1

def decompress_bc1_block(block:bytes, alpha:bool=True):
    # First 2+2 bytes: color0 and color1 (16-bit RGB565)
    c0, c1 = struct.unpack('<HH', block[0:4])
    # Decode colors into RGBA format
    colors = [rgb565_to_rgba(c0), rgb565_to_rgba(c1)]
    # Intermediate colors
    if c0 > c1 or not alpha:
        colors.append((colors[0] * 2 + colors[1]) / 3)
        colors.append((colors[0] + colors[1] * 2) / 3)
    else:
        colors.append((colors[0] + colors[1]) / 2)
        colors.append(np.array([0, 0, 0, 0]))
    
    # Next 4 bytes: pixel indices (2 bits per pixel for 4x4 block)
    indices = struct.unpack('<L', block[4:8])[0]
    pixel_data = []
    for bit_offset in range(0, 32, 2):
        pixel_index = (indices >> bit_offset) & 0x03
        pixel_data.append(colors[pixel_index])

    return np.array(pixel_data, dtype=np.uint8).reshape((4, 4, 4))


def decompress_bc2_block(block:bytes):
    # BC2 block contains the same color data as BC1, but with an additional 8-byte alpha block
    alpha_block, color_block = block[0:8], block[8:16]
    
    # 8 bytes (4 bits per pixel for 4x4 block)
    alpha_bits = struct.unpack("<Q", alpha_block)[0]
    alpha_data =  [((alpha_bits >> i) & 0x0F) << 4 for i in range(0, 64, 4)]
    
    # Apply alpha data to color block
    pixel_data = decompress_bc1_block(color_block, alpha=False)
    pixel_data[:, :, 3] = np.array(alpha_data, dtype=np.uint8).reshape((4, 4))
    return pixel_data


def decompress_bc3_block(block:bytes):
    # BC3 block contains the same color data as BC1, but with an additional 8-byte alpha block
    alpha_block, color_block = block[0:8], block[8:16]
    
    # First 1+1 bytes: alpha1 and alpha2 (8-bit)
    a0, a1 = struct.unpack('<BB', alpha_block[0:2])
    alphas = [a0, a1]
    # Intermediate alphas
    if a0 > a1:
        for i in range(1, 7):
            alphas.append((a0 * (7 - i) + a1 * i) // 7)
    else:
        for i in range(1, 5):
            alphas.append((a0 * (5 - i) + a1 * i) // 5)
        alphas.extend([0, 255])

    # Next 6 bytes: pixel indices (3 bits per pixel for 4x4 block)
    ## Pad with 2 zero bytes to make it 8 bytes
    indices = struct.unpack('<Q', alpha_block[2:8] + b'\x00\x00')[0] 
    alpha_data = []
    for bit_offset in range(0, 48, 3):
        pixel_index = (indices >> bit_offset) & 0x07
        alpha_data.append(alphas[pixel_index])

    # Apply alpha data to color block
    pixel_data = decompress_bc1_block(color_block, alpha=False)
    pixel_data[:, :, 3] = np.array(alpha_data, dtype=np.uint8).reshape((4, 4))
    return pixel_data


# %%
class DtxImage:
    
    def _read_bpp_8p(self):
        raise NotImplementedError
        
    def _read_bpp_8(self):
        for x in range(self.height):
            for y in range(self.width):
                pixel_bits = struct.unpack("<B", self.fStream.read(1))[0]
                b, g, r, a =  [((pixel_bits >> i) & 0x03) << 2 for i in range(0, 8, 2)]
                self._pixels[x, y] = [r, g, b, a]
        
    def _read_bpp_16(self):
        for x in range(self.height):
            for y in range(self.width):
                pixel_bits = struct.unpack("<H", self.fStream.read(2))[0]
                b, g, r, a =  [((pixel_bits >> i) & 0x0F) << 4 for i in range(0, 16, 4)]
                self._pixels[x, y] = [r, g, b, a]
        
    def _read_bpp_32(self):
        for x in range(self.height):
            for y in range(self.width):
                b, g, r, a = struct.unpack('<BBBB', self.fStream.read(4))
                self._pixels[x, y] = [r, g, b, a]

    def _read_bpp_s3tc_dxt1(self):
        block_size = 8  # Each BC1 block is 8 bytes
        block_dim = 4 # Each BC1 block is 4x4 pixels
        for block_y in range(0, self.height, block_dim):
            for block_x in range(0, self.width, block_dim):
                # Calculate the offset of the current block in the texture data
                block = self.fStream.read(block_size)
                # Decompress the block
                self._pixels[
                    block_y: block_y + block_dim, 
                    block_x: block_x + block_dim
                ] = decompress_bc1_block(block)
        
    def _read_bpp_s3tc_dxt3(self):
        block_size = 16  # Each BC2 block is 8 bytes
        block_dim = 4 # Each BC2 block is 4x4 pixels
        for block_y in range(0, self.height, block_dim):
            for block_x in range(0, self.width, block_dim):
                # Calculate the offset of the current block in the texture data
                block = self.fStream.read(block_size)
                # Decompress the block
                self._pixels[
                    block_y: block_y + block_dim, 
                    block_x: block_x + block_dim
                ] = decompress_bc2_block(block)
        
    def _read_bpp_s3tc_dxt5(self):
        block_size = 16  # Each BC3 block is 8 bytes
        block_dim = 4 # Each BC3 block is 4x4 pixels
        for block_y in range(0, self.height, block_dim):
            for block_x in range(0, self.width, block_dim):
                # Calculate the offset of the current block in the texture data
                block = self.fStream.read(block_size)
                # Decompress the block
                self._pixels[
                    block_y: block_y + block_dim, 
                    block_x: block_x + block_dim
                ] = decompress_bc3_block(block)
        
    def _read_bpp_32p(self):
        raise NotImplementedError
        
    def _read_bpp_24(self):
        for x in range(self.height):
            for y in range(self.width):
                b, g, r = struct.unpack('<BBB', self.fStream.read(3))
                self._pixels[x, y] = [r, g, b, 255]
        
        
    def __init__(self, path: str):
        # Maybe LZMA compressed, try decompressing
        if open(path, "rb").read(4) == b"\x5D\x00\x00\x00":
            self.fStream = lzma.open(path, "rb")
        else:
            self.fStream = open(path, "rb")

        self.m_Header = DtxHeader(self.fStream.read(164))
        
        self.width = self.m_Header.m_BaseWidth
        self.height = self.m_Header.m_BaseHeight
        
        # Safely check version
        if self.m_Header.m_Version != CURRENT_DTX_VERSION:
            raise NotImplementedError("Unsupported DTX version: {}".format(self.m_Header.m_Version))


        if self.m_Header.m_nMipmaps > 1:
            print("Warning: DTX file has multiple mipmaps, only the first mip (highest resolution) will be loaded!")

        self._pixels = np.zeros((self.height, self.width, 4), dtype=np.uint8)

        match BPPIdent(self.m_Header.get_bpp_ident()):
            case BPPIdent.BPP_8P:
                self._read_bpp_8p()
            case BPPIdent.BPP_8:
                self._read_bpp_8()
            case BPPIdent.BPP_16:
                self._read_bpp_16()
            case BPPIdent.BPP_32:
                self._read_bpp_32()
            case BPPIdent.BPP_S3TC_DXT1:
                self._read_bpp_s3tc_dxt1()
            case BPPIdent.BPP_S3TC_DXT3:
                self._read_bpp_s3tc_dxt3()
            case BPPIdent.BPP_S3TC_DXT5:
                self._read_bpp_s3tc_dxt5()
            case BPPIdent.BPP_32P:
                self._read_bpp_32p()
            case BPPIdent.BPP_24:
                self._read_bpp_24()
            case _:
                raise NotImplementedError("Unsupported BPP ident: {}".format(BPPIdent(self.m_Header.get_bpp_ident())))
        
        self.fStream.close()
        self.image = Image.fromarray(self._pixels)

# %%

# %%
def process_dtx(filename):
    try:
        dtx = DtxImage(filename)
        dtx.image.save(filename.rsplit(".", maxsplit=1)[0] + ".png")
    except KeyboardInterrupt:
        raise
    except Exception as e:
        print(f"Error processing {filename}: {e}")
        return False
    else:
        os.remove(filename)
        return True

if __name__ == '__main__':

    # dtx_files = glob(r"dump/rf019/**/*.DTX", recursive=True)

    # pool = Pool(processes=16)
    # for _ in tqdm.tqdm(pool.imap(process_dtx, dtx_files), total=len(dtx_files)):
    #     pass
    
    process_dtx(sys.argv[1])
